import path from 'path';
import { createRequire } from 'module';
import _ from 'lodash';
import _debug from 'debug';
import Version from './version/Version.js';
import Git from './git/Git.js';
import GitLab from './gitlab/GitLab.js';
import GitHub from './github/GitHub.js';
import npm from './npm/npm.js';

const require = createRequire(import.meta.url);

const debug = _debug('release-it:plugins');

const pluginNames = ['npm', 'git', 'github', 'gitlab', 'version'];

const plugins = {
  version: Version,
  git: Git,
  gitlab: GitLab,
  github: GitHub,
  npm: npm
};

const load = async pluginName => {
  let plugin = null;
  try {
    const module = await import(require.resolve(pluginName));
    plugin = module.default;
  } catch (err) {
    const module = await import(require.resolve(path.resolve(pluginName)));
    plugin = module.default;
  }
  return [path.parse(pluginName).name, plugin];
};

export let getPlugins = async (config, container) => {
  const context = config.getContext();
  const disabledPlugins = [];

  const enabledExternalPlugins = await _.reduce(
    context.plugins,
    async (result, pluginConfig, pluginName) => {
      const [name, Plugin] = await load(pluginName);
      const [namespace, options] = pluginConfig.length === 2 ? pluginConfig : [name, pluginConfig];
      config.setContext({ [namespace]: options });
      if (await Plugin.isEnabled(options)) {
        const instance = new Plugin({ namespace, options: config.getContext(), container });
        debug({ namespace, options: instance.options });
        (await result).push(instance);
        disabledPlugins.push(..._.intersection(pluginNames, _.castArray(Plugin.disablePlugin(options))));
      }
      return result;
    },
    []
  );

  const enabledPlugins = await pluginNames.reduce(async (result, plugin) => {
    if (plugin in plugins && !disabledPlugins.includes(plugin)) {
      const Plugin = plugins[plugin];
      const pluginOptions = context[plugin];
      if (await Plugin.isEnabled(pluginOptions)) {
        const instance = new Plugin({ namespace: plugin, options: context, container });
        debug({ namespace: plugin, options: instance.options });
        (await result).push(instance);
      }
    }
    return result;
  }, []);

  return [enabledPlugins, enabledExternalPlugins];
};
